//   HBF2FNT.c
//   converts "HBFable" hanzi bitmap font file
//         to MS Windows .FNT font data files
//
//   compile and link using MS VC++ 1.0 or MS C 7.0:
//     cl /AC hbf2fnt.c hbfapi.c
//
//   Tian Bogang,   November 1993

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#define DWORD unsigned long
#define WORD  unsigned int
#define BYTE  unsigned char

#define FIRST_CHAR 0x20
#define LAST_CHAR  0xFF
#define NUM_CHAR   (LAST_CHAR-FIRST_CHAR+1)

#define BYN 0x1C1

// -------- .FNT file header --------
//
// see <<Programmer's Reference, Volume 4: Resource>>, p49
//     in MS Windows 3.1 SDK
//
WORD  dfVersion = 0x0200;
DWORD dfSize;
char  dfCopyright [61] =
      {"MS Windows font converted from public domain hanzi bitmap.   "};
WORD  dfType = 0;
WORD  dfPoints = 72;
WORD  dfVertRes;
WORD  dfHorizRes;
WORD  dfAscent;
WORD  dfInternalLeading = 0;
WORD  dfExternalLeading = 0;
BYTE  dfItalic = 0;
BYTE  dfUnderline = 0;
BYTE  dfStrikeOut = 0;
WORD  dfWeight = 400;
BYTE  dfCharSet = 3;
WORD  dfPixWidth;
WORD  dfPixHeight;
BYTE  dfPitchAndFamily = 0;
WORD  dfAvgWidth;
WORD  dfMaxWidth;
BYTE  dfFirstChar = FIRST_CHAR;
BYTE  dfLastChar = LAST_CHAR;
BYTE  dfDefaultChar = 0x60;
BYTE  dfBreakChar = 0;
WORD  dfWidthBytes = BYN;
DWORD dfDevice = 0;
DWORD dfFace;
DWORD dfBitsPointer = 0;
DWORD dfBitsOffset = (NUM_CHAR + 1)*4 + 118;
BYTE  dfReserved = 0;
WORD  dfCharTable [2*NUM_CHAR + 2];
char  FaceName [10];
//
// -- end of .FNT file header ------------------

char hbffile [14], fontname [14], FNTname [14];
int  font_code, num_read, NUM_BYTE, BYTE_PER_LINE;
BYTE *font_bitmap [NUM_CHAR];
short SideWay;

extern int mx, my, yd, CHsize;
extern BYTE *buffer;

int  BuildCharTable ();
int  BuildFontHeader ();
int  GetBitmap ();
int  PutFaceName (FILE *);
int  PutFontBitmap (FILE *);
int  PutFontHeader (FILE *);
void TransposeBitmap (char *, int);

main (int argc, char **argv)
{
  FILE *fout;
  char buf [10];
  int  i;

  if (argc > 2) {
     strcpy (hbffile, argv [1]);
     strcpy (fontname, argv [2]);
  }
  else {
    printf ("\nusage : hbf2fnt hbffile fontfile");
    exit (-1);
  }

  if (HBFopen (hbffile) != 1) {
    printf ("\nerror reading HBF file %s", hbffile);
    exit (-1);
  }

  for (i = 0; i < NUM_CHAR; i++)
    if ((font_bitmap [i] = (BYTE *) calloc (CHsize, sizeof(BYTE))) == NULL) {
      printf ("\nnot enough memory for font bitmap!");
      exit (-1);
    }

  BuildFontHeader ();

  font_code = 0;
  while (GetBitmap ()) {
    strcpy (FaceName, fontname);
    strcpy (FNTname, fontname);
    sprintf (buf, "_%02d", (font_code + 1));
    strcat (FNTname, buf);
    strcat (FaceName, buf);
    strcat (FNTname, ".FNT");
    if ((fout = fopen (FNTname, "wb")) == NULL) {
      printf ("\nerror open .FNT file %s", FNTname);
      exit (-1);
    }
    BuildCharTable ();
    PutFontHeader (fout);
    PutFontBitmap (fout);
    PutFaceName (fout);
    fclose (fout);
    font_code++;
    printf ("\n %d hanzi bitmaps ==> %s", num_read, FNTname);
  }

  for (i = 0; i < NUM_CHAR; i++) free (font_bitmap [i]);
  HBFclose ();
}

int BuildFontHeader (void)
{
  NUM_BYTE = CHsize;
  dfPixWidth = dfAvgWidth = dfMaxWidth = dfHorizRes = mx;
  dfPixHeight = dfVertRes = my;
  dfAscent = my + yd ;
  BYTE_PER_LINE = (mx+7)/8;
  dfFace = dfBitsOffset + NUM_CHAR*NUM_BYTE + dfPixHeight;
  dfSize = dfFace + strlen (fontname) + 4;

  return 1;
}

int BuildCharTable (void)
{
  int i;

  for (i = 0; i < NUM_CHAR; i++) {
    dfCharTable [i*2] = dfPixWidth;
    dfCharTable [i*2 + 1] = (WORD) (dfBitsOffset + i*NUM_BYTE);
  }
  // there is one extra entry at the end, the "absolute-space"
  // its width is always set to 8
  dfCharTable [NUM_CHAR*2] = 8;
  dfCharTable [NUM_CHAR*2 + 1] = (WORD) (dfBitsOffset + NUM_CHAR*NUM_BYTE);

  return 1;
}

int GetBitmap ()
{
  int l, i, j, k;

  num_read = 0;
  for (j = 0; j < NUM_CHAR; j++) {
    if (HBFgetNextBitmap () == 0) {  // get the bitmap of a hanzi
      if (j == 0) return 0;
      break;
    }
    if (SideWay) TransposeBitmap (buffer, NUM_BYTE);
    num_read++;

    // The bitmap of Windows raster font is arranged
    // in a different way from hanzi bitmap.
    // In Windows bitmap, "the first byte contains the first
    // 8 bits of the scan line. The second byte contains the
    // first 8 bits of the second scan line. This continues
    // until the first "column" is completely defined. The
    // subsequent byte contains the next 8 bits of the first
    // scan line, ..."
    // from <<Programmer's Reference, Volume 4: Resource>>, p55-56
    //      in MS Windows 3.1 SDK
    //
    // Take the 24*24 hanzi bitmap as an example
    //
    //   == this is hanzi bitmap ==   |   == this is Windows bitmap ==
    //   -byte0-   -byte1-   -byte2-  |   -byte0-   -byte24-  -byte48-
    //   -byte3-   -byte4-   -byte5-  |   -byte1-   -byte25-  -byte49-
    //   ...........................  |   ...........................
    //   -byte69-  -byte70-  -byte71- |   -byte23-  -byte47-  -byte71-
    //
    // the conversion is done in the following

    for (l = 0; l < (dfPixWidth / 8); l++) {
      for (i = (l * dfPixWidth); i < ((l + 1) * dfPixWidth); i++) {
        k = (i - (l * dfPixWidth)) * (dfPixWidth / 8) + l;
        font_bitmap [j] [i] = buffer [k];
      }
    }

  }   // end of for j

  // null bitmaps
  if (num_read < NUM_CHAR)
    for (j = num_read; j < NUM_CHAR; j++)
      for (i = 0; i < NUM_BYTE; i++)
        font_bitmap [j] [i] = 0;

  return 1;
}

int PutFaceName (FILE *fp)
{
  fprintf (fp, "%s", FaceName);
  fprintf (fp, "%c", '\0');

  return 1;
}

int PutFontBitmap (FILE *fp)
{
  int i, j;

  for (i = 0; i < NUM_CHAR; i++)
    for (j = 0; j < NUM_BYTE; j++)
      if (fprintf (fp, "%c", font_bitmap [i] [j]) < 0) {
        printf ("\nerror putting font bitmap %d %d!", i, j);
        exit (-1);
      }

  // there is one extra entry at the end, the "absolute-space"
  for (j = 0; j < dfPixHeight; j++)
    if (fprintf (fp, "%c", '\0') < 0) {
      printf ("\nerror putting null bitmap %d!", j);
      exit (-1);
    }
}

int PutFontHeader (FILE *fp)
{
  fwrite ((WORD *) &dfVersion, sizeof(WORD), 1, fp);
  fwrite ((DWORD *) &dfSize, sizeof(DWORD), 1, fp);
  fwrite ((char *) dfCopyright, sizeof(char), 60, fp);
  fwrite ((WORD *) &dfType, sizeof(WORD), 1, fp);
  fwrite ((WORD *) &dfPoints, sizeof(WORD), 1, fp);
  fwrite ((WORD *) &dfVertRes, sizeof(WORD), 1, fp);
  fwrite ((WORD *) &dfHorizRes, sizeof(WORD), 1, fp);
  fwrite ((WORD *) &dfAscent, sizeof(WORD), 1, fp);
  fwrite ((WORD *) &dfInternalLeading, sizeof(WORD), 1, fp);
  fwrite ((WORD *) &dfExternalLeading, sizeof(WORD), 1, fp);
  fwrite ((BYTE *) &dfItalic, sizeof(BYTE), 1, fp);
  fwrite ((BYTE *) &dfUnderline, sizeof(BYTE), 1, fp);
  fwrite ((BYTE *) &dfStrikeOut, sizeof(BYTE), 1, fp);
  fwrite ((WORD *) &dfWeight, sizeof(WORD), 1, fp);
  fwrite ((BYTE *) &dfCharSet, sizeof(BYTE), 1, fp);
  fwrite ((WORD *) &dfPixWidth, sizeof(WORD), 1, fp);
  fwrite ((WORD *) &dfPixHeight, sizeof(WORD), 1, fp);
  fwrite ((BYTE *) &dfPitchAndFamily, sizeof(BYTE), 1, fp);
  fwrite ((WORD *) &dfAvgWidth, sizeof(WORD), 1, fp);
  fwrite ((WORD *) &dfMaxWidth, sizeof(WORD), 1, fp);
  fwrite ((BYTE *) &dfFirstChar, sizeof(BYTE), 1, fp);
  fwrite ((BYTE *) &dfLastChar, sizeof(BYTE), 1, fp);
  fwrite ((BYTE *) &dfDefaultChar, sizeof(BYTE), 1, fp);
  fwrite ((BYTE *) &dfBreakChar, sizeof(BYTE), 1, fp);
  fwrite ((WORD *) &dfWidthBytes, sizeof(WORD), 1, fp);
  fwrite ((DWORD *) &dfDevice, sizeof(DWORD), 1, fp);
  fwrite ((DWORD *) &dfFace, sizeof(DWORD), 1, fp);
  fwrite ((DWORD *) &dfBitsPointer, sizeof(DWORD), 1, fp);
  fwrite ((DWORD *) &dfBitsOffset, sizeof(DWORD), 1, fp);
  fwrite ((BYTE *) &dfReserved, sizeof(BYTE), 1, fp);
  fwrite ((WORD *) dfCharTable, sizeof(WORD), (2*(NUM_CHAR + 1)), fp);

  return 1;
}

/* ----- This routine transposes a chinese bitmap ----- */
void TransposeBitmap (char *bmp_buf, int num_of_byte)
/* bmp_buf       is the input bitmap to be transposed    */
/* num_of_byte   is the number of byte of the bitmap     */
/* the hanzi bitmap is supposed to be square             */
{
  int i, j, k, l, J, l1, k1, i1, i0;
  unsigned char temp;
  char *bmp1_buf;

  if ((bmp1_buf = (char *) calloc (num_of_byte+1, sizeof(char))) == NULL) {
    printf ("\nnot enough memory for transposed bitmap!");
    exit (-1);
  }

  for (k = 0; k < num_of_byte; k++) {
    bmp1_buf [k] = 0;
  }

  for (k = 0; k < num_of_byte; k++) {
    /* i is the row of a dot, j is the byte which contains  */
    /* the the dot. The relation is                         */
    /*    i = k / byte_per_line                             */
    /*    j = k - (i * byte_per_line)                       */
    /*      -or-                                            */
    /*    k = (i * 3) + j                                   */

    i = k / BYTE_PER_LINE;  j = k - (i * BYTE_PER_LINE);
    i0 = i / 8;  l1 = i % 8;
    for (l = 0; l < 8; l++) {
      /*  l is the bit position of dot in a byte, l = 0 to 7  */
      /*  J is the column of the dot                          */
      /*  Relation :                                          */
      /*      J = (j * 8) + l,  j = J / 8,  l = J % 8         */
      /*  In transpose (i, J) --> (J, i) = (i1, j1)           */
      /*  so k1 = (i1 * 3) + j1 = (J * 3) + i / 8             */
      /*     l1 = i % 8                                       */
      /*  so the l'th dot of k'th byte in bmp_buf is put      */
      /*  at the l1'th dot of k1'th byte of bmp_buf1          */
      /*  note that l'th dot = (7 - l)'th bit in a byte       */

      J = (j * 8) + l;
      k1 = (J * BYTE_PER_LINE) + i0;
      temp = 0;
      if (((bmp_buf [k] >> (7 - l)) & 1) != 0) temp = 1;
      bmp1_buf [k1] |= (temp << (7-l1));
    }
  }

  /* now copy the transposed bitmap back to bmp_buf */
  for (k = 0; k < num_of_byte; k++) {
    bmp_buf [k] = bmp1_buf [k];
  }
  free (bmp1_buf);
}
